*     **** QUASI-SECANT METHOD  ****
*
*     A Solver for Nonsmooth Optimization
*
*     by Adil Bagirov
*
*     Centre for Informatics and Applied Optimization
*     ITMS, University of Ballarat
*
*
C     Last change:  AB    3 Dec 2008    5:31 pm

      implicit double precision (a-h,o-z)
      PARAMETER(maxvar=2000)
      double precision x0(maxvar),x(maxvar),t(21)
      COMMON /cnf/nf,/cngrad2/ngrad,/ct/t
      character*30 outfi/'results.txt'/
      open(40,file=outfi)
c=======================================================
c  Input data:
c  n        - number of variables
c=======================================================
      n  = 4

      do i=1,21
       t(i)=-5.d-01+DBLE(i-1)/2.d+01
      end do 
c======================================================
c nbundle - is the size of bundle, the maximum number of
c subgradients or discrete gradients which can be
c calculated at each iteration
c======================================================
      nbundle=21
c======================================================
c ngradient = 0 if you use approximation to subgradient
c ngradient = 1 if you use exact gradients of the objective
c               and constraint functions
c======================================================
      ngradient=1
c=======================================================
c  Starting point
c=======================================================
      x0(1)=1.0d+00
      x0(2)=1.0d+00
      x0(3)=-3.0d+00
      x0(4)=-1.0d+00
c=======================================================
c  Calling DGM
c=======================================================
      call dgm(n,x0,nbundle,ngradient,x,fvalue,cputime,finit)
      WRITE(40,41) finit,fvalue,nf,ngrad,cputime
  41  FORMAT(2f16.8,2i8,f10.4)
c=======================================================
      CLOSE(40)
      STOP
      END

c======================================================================
      subroutine func(x,objf)
      implicit double precision (a-h,o-z)
      PARAMETER(maxvar=2000)
      double precision x(maxvar),t(21),fi(21)
      COMMON /cka/ka,ind,/ct/t
c===============================================================
      do i=1,21
        fi(i)=x(1)*dexp(x(3)*t(i))+x(2)*dexp(x(4)*t(i))
     1   -1.d+00/(1.d+00+t(i))
      end do

      f0=dabs(fi(1))

      do i=2,21
       f0=dmax1(f0,dabs(fi(i)))
      end do

      objf=f0

      do i=1,21
       IF(dabs(fi(i)).eq.objf) ka=i
      end do

      if(fi(ka).ge.0.d+00) ind=1
      if(fi(ka).lt.0.d+00) ind=-1
c======================================================
      return
      end

c======================================================================
      subroutine gradient(x,grad)
      implicit double precision (a-h,o-z)
      PARAMETER(maxvar=2000)
      double precision x(maxvar),grad(maxvar),t(21)
      COMMON /cka/ka,ind,/ct/t
c===============================================================
      grad(1)=dexp(x(3)*t(ka))
      grad(2)=dexp(x(4)*t(ka))
      grad(3)=x(1)*t(ka)*dexp(x(3)*t(ka))
      grad(4)=x(2)*t(ka)*dexp(x(4)*t(ka))
      do i=1,4
       grad(i)=ind*grad(i)
      end do
c===============================================================
      return
      end

c======================================================================
c Without scaling
c======================================================================
      subroutine dgm(nvar,x0,nbundle,ngradient,x,f,cputime,finit)
      implicit double precision (a-h,o-z)
      PARAMETER(maxvar=2000)
      double precision x(maxvar),x0(maxvar)
      COMMON /csize/m,/citer/maxiter,niter,/cnf/nf,/cgrad/ngradient1
     1 ,/cngrad2/ngrad2
      call cpu_time(time1)
      m=nvar
      nf=0
      niter=0
      ngrad2=0
      ngradient1=ngradient
      maxiter=10000
      slinit=1.d+00
      do i=1,m
        x(i)=x0(i)
      end do
      call fv(x,f)
      finit=f
      call optimum(x,nbundle,slinit)
      call fv(x,f)
      call cpu_time(time2)
      cputime=time2-time1
c===========================================================
      return
      end

c============================================================
      subroutine optimum(x,nbundle,slinit)
      implicit double precision (a-h,o-z)
      PARAMETER(maxvar=2000, maxdg=1000,maxit=100000)
      double precision x(maxvar),x1(maxvar),g(maxvar),v(maxvar)
     1 ,w(maxdg,maxvar),prod(maxdg,maxdg),z(maxdg),fvalues(maxit)
      INTEGER ij(maxdg)
      common /csize/m,/citer/maxiter,niter,/cij/ij,jvertex,/cz/z
     1 ,/ckmin/kmin
c====================================================================
      dist1=1.d-07
      step0=-2.d-01
      div=4.d-01
      eps0=1.d-07
      slmin=1.d-10*slinit
      pwt=1.0d-08
      sdif=1.d-05
      mturn=4
c====================================================================
      sl=slinit
      call fv(x,f2)
  1   sl=div*sl
      IF(sl.lt.slmin) return
      IF(sl.le.1.d-05) step0=-9.d-01
      do i=1,m
       g(i)=1.d+00/dsqrt(DBLE(m))
      end do
      nnew=0
c================================================================
   2  niter=niter+1
        IF(niter.gt.maxiter) RETURN
        nnew=nnew+1
        f1=f2
        fvalues(niter)=f1
c---------------------------------------------------------------
        if (nnew.gt.mturn) then
         mturn2=niter-mturn+1
         ratio1=(fvalues(mturn2)-f1)/(dabs(f1)+1.d+00)
         IF(ratio1.LT.sdif) GO TO 1
        end if
        if (nnew.GE.(2*mturn)) then
         mturn2=niter-2*mturn+1
         ratio1=(fvalues(mturn2)-f1)/(dabs(f1)+1.d+00)
         IF(ratio1.LT.(1.d-01*sdif)) GO TO 1
        end if
c--------------------------------------------------------------
        do ndg=1,nbundle
            call dgrad(x,sl,g,v,f1,f4,ndg,pwt)
            dotprod=0.d+00
            do i=1,m
             dotprod=dotprod+v(i)*v(i)
            end do
            r=dsqrt(dotprod)
            IF(r.lt.eps0) GO TO 1
            IF(ndg.eq.1) then
                         rmean=r
                         kmin=1
                         rmin=r
            END if
            IF(ndg.gt.1) then
                         rmin=dmin1(rmin,r)
                         IF(r.eq.rmin) kmin=ndg
                         rmean=((ndg-1)*rmean+r)/ndg
            END if
            toler=dmax1(eps0,dist1*rmean)
            do i=1,ndg-1
             prod(ndg,i)=0.d+00
             do j=1,m
              prod(ndg,i)=prod(ndg,i)+w(i,j)*v(j)
             end do
             prod(i,ndg)=prod(ndg,i)
            end do
            prod(ndg,ndg)=dotprod
c====================================================================
            do i=1,m
             w(ndg,i)=v(i)
            end do
            call wolfe(ndg,prod)
c================================
            do i=1,m
             v(i)=0.d+00
             do j=1,jvertex
              v(i)=v(i)+w(ij(j),i)*z(j)
             END do
            END do
c================================
            r=0.d+00
            do i=1,m
             r=r+v(i)*v(i)
            end do
            r=dsqrt(r)
            if(r.lt.toler) GO TO 1
c===========================================================
             do i=1,m
              g(i)=-v(i)/r
              x1(i)=x(i)+sl*g(i)
             end do
c===========================================================
             call fv(x1,f4)
             f3=(f4-f1)/sl
             decreas=step0*r
             if(f3.lt.decreas) then
                        call armijo(x,g,f1,f5,f4,sl,step,r)
                        f2=f5
                        do i=1,m
                         x(i)=x(i)+step*g(i)
                        end do
c                        print 3, niter,ndg,f2,sl,r
                        IF(ndg.le.2) sl=1.1d+00*sl
                        GO TO 2
 3                      format(I6,I4,3f16.8)
             end if
         END do
c=====================================================
      go to 1
      return
      end

c==============================================================
c  Subroutines Wolfe and Equations solves quadratic
c  programming problem, to find
c  descent direction, Step 3, Algorithm 2.
c===============================================================

      subroutine wolfe(ndg,prod)
      implicit double precision (a-h,o-z)
      PARAMETER(maxvar=2000, maxdg=1000)
      common /csize/m,/w01/a,/cij/ij,jvertex,/cz/z,/ckmin/kmin
      INTEGER ij(maxdg)
      double precision z(maxdg),z1(maxdg),a(maxdg,maxdg)
     1 ,prod(maxdg,maxdg)
      j9=0
      jmax=500*ndg
      jvertex=1
      ij(1)=kmin
      z(1)=1.d+00
c=======================================
c  To calculate X
c=======================================
 1    r=0.d+00
      do i=1,jvertex
       do j=1,jvertex
        r=r+z(i)*z(j)*prod(ij(i),ij(j))
       end do
      end do
      IF(ndg.eq.1) RETURN
c========================================
c  To calculate <X,P_J> and J
c========================================
      t0=1.d+12
      do i=1,ndg
        t1=0.d+00
        do j=1,jvertex
          t1=t1+z(j)*prod(ij(j),i)
        end do
        if(t1.lt.t0) then
                     t0=t1
                     kmax=i
        end if
      end do
c========================================
c  First stopping criterion
c========================================
      rm=prod(kmax,kmax)
      do j=1,jvertex
       rm=dmax1(rm,prod(ij(j),ij(j)))
      end do
      r2=r-1.d-12*rm
      if(t0.gt.r2) RETURN
c========================================
c  Second stopping criterion
c========================================
      do i=1,jvertex
       if(kmax.eq.ij(i)) RETURN
      end do
c========================================
c Step 1(e) from Wolfe's algorithm
c========================================
      jvertex=jvertex+1
      ij(jvertex)=kmax
      z(jvertex)=0.d+00
c========================================
 2    do i=1,jvertex
       do j=1,jvertex
        a(i,j)=1.d+00+prod(ij(i),ij(j))
       end do
      end do
      j9=j9+1
      if(j9.gt.jmax) RETURN
      call equations(jvertex,z1)
      do i=1,jvertex
       if(z1(i).le.1.d-10) go to 3
      end do
      do i=1,jvertex
       z(i)=z1(i)
      end do
      go to 1
  3   teta=1.d+00
      do i=1,jvertex
       z5=z(i)-z1(i)
       if(z5.gt.1.d-10) teta=dmin1(teta,z(i)/z5)
      end do
      do i=1,jvertex
       z(i)=(1.d+00-teta)*z(i)+teta*z1(i)
       if(z(i).le.1.d-10) then
                          z(i)=0.d+00
                          kzero=i
       end if
      end do
      j2=0
      do i=1,jvertex
       IF(i.ne.kzero) then
                     j2=j2+1
                     ij(j2)=ij(i)
                     z(j2)=z(i)
       END if
      end do
      jvertex=j2
      go to 2
      return
      end

      subroutine equations(n,z1)
      implicit double precision (a-h,o-z)
      PARAMETER(maxvar=2000, maxdg=1000)
      common /w01/a
      double precision a(maxdg,maxdg),z1(maxdg),b(maxdg,maxdg)
      do i=1,n
       do j=1,n
        b(i,j)=a(i,j)
       end do
       b(i,n+1)=1.d+00
      end do
      do i=1,n
       r=b(i,i)
       do j=i,n+1
        b(i,j)=b(i,j)/r
       end do
       do j=i+1,n
        do k=i+1,n+1
         b(j,k)=b(j,k)-b(i,k)*b(j,i)
        end do
       end do
      end do
      z1(n)=b(n,n+1)
      do i=1,n-1
        k=n-i
        z1(k)=b(k,n+1)
        do j=k+1,n
         z1(k)=z1(k)-b(k,j)*z1(j)
        END do
      end do
      z2=0.d+00
      do i=1,n
       z2=z2+z1(i)
      end do
      do i=1,n
       z1(i)=z1(i)/z2
      end do
      return
      end

c=====================================================================
c Subroutine dgrad calculates subgradients or discrete gradients
c=====================================================================
      subroutine dgrad(x,sl,g,dg,f1,f4,ndg,pwt)
      implicit double precision (a-h,o-z)
      PARAMETER(maxvar=2000, maxdg=1000)
      double precision x1(maxvar),g(maxvar),x(maxvar),dg(maxvar)
      common /csize/m,/cgrad/ngradient1,/cngrad2/ngrad2

      do k=1,m
        x1(k)=x(k)+sl*g(k)
      end do

      if (ngradient1.eq.0) then
       IF(ndg.gt.1) r2=f4
       IF(ndg.eq.1) call fv(x1,r2)
       call dgrad2(x1,dg,r2,pwt)
      end if

      if (ngradient1.eq.1) then
       call gradient(x1,dg)
       ngrad2=ngrad2+1
      end if

      return
      end

c=====================================================================
c Subroutine dgrad calculates discrete gradients: Step 5, Algorithm 1
c=====================================================================
      subroutine dgrad2(x1,v,r2,pwt)
      implicit double precision (a-h,o-z)
      PARAMETER(maxvar=2000, maxdg=1000)
      double precision x1(maxvar),v(maxvar)
      common /csize/m
      alpha=1.d+00
      t=pwt/alpha
      r4=r2
      do k=1,m
        t=t*alpha
        r3=r4
        x1(k)=x1(k)+t
        call fv(x1,r4)
        v(k)=(r4-r3)/t
      end do
      return
      end

c===========================================================
c Line search (Armijo-type), Step 5 Algorithm 2.
c===========================================================
      subroutine armijo(x,g,f1,f5,f4,sl,step,r)
      implicit double precision (a-h,o-z)
      PARAMETER(maxvar=2000, maxdg=1000)
      common /csize/m
      double precision x(maxvar),g(maxvar),x1(maxvar)
      step=5.d+01*r
  1   step=5.d-01*step
      if (step.le.sl) then
       step=sl
       f5=f4
       return
      end if
      do i=1,m
       x1(i)=x(i)+step*g(i)
      end do
      call fv(x1,f5)
      f3=f5-f1+1.d-04*step*r
      IF(f3.le.0.d+00) RETURN
      GO TO 1
      return
      end

c===========================================================
c Line search (Armijo-type), Step 5 Algorithm 2.
c===========================================================
c      subroutine armijo(x,g,f1,f5,f4,sl,step,r)
c      implicit double precision (a-h,o-z)
c      PARAMETER(maxvar=2000, maxdg=1000)
c      common /csize/m
c      double precision x(maxvar),g(maxvar),x1(maxvar),fline(1000)
c     1 ,sline(1000)
c      k=1
c      fline(k)=f4
c      sline(k)=sl
c      step=sl
c  1   step=2*step
c      do i=1,m
c       x1(i)=x(i)+step*g(i)
c      end do
c      call fv(x1,f5)
c      f3=f5-f1+1.d-04*step*r
c      IF(f3.gt.0.d+00) GO TO 3
c      IF(f3.le.0.d+00) GO TO 2
c 2    k=k+1
c      fline(k)=f5
c      sline(k)=step
c      IF(k.gt.50) GO TO 3
c      GO TO 1
c 3    continue
c      step=sline(1)
c      f5=fline(1)
c      do i=2,k
c       if (f5.gt.fline(i)) then
c          f5=fline(i)
c          step=sline(i)
c       end if
c      end do
c      return
c      end

      subroutine fv(x,f)
      implicit double precision (a-h,o-z)
      PARAMETER(maxvar=2000)
      double precision x(maxvar)
      COMMON /cnf/nf,/csize/m
c================================================================
      call func(x,objf)
      nf=nf+1
      f=objf
c=================================================================
      return
      end
